# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Networking
----------

The ``drgn.helpers.linux.net`` module provides helpers for working with the
Linux kernel networking subsystem.
"""

import ipaddress
import operator
from typing import Iterator, List, Optional, Union

from drgn import NULL, IntegerLike, Object, Program, Type, cast, container_of, sizeof
from drgn.helpers.common.prog import (
    takes_object_or_program_or_default,
    takes_program_or_default,
)
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.list import hlist_for_each_entry, list_for_each_entry
from drgn.helpers.linux.list_nulls import hlist_nulls_for_each_entry

__all__ = (
    "SOCK_INODE",
    "SOCKET_I",
    "for_each_net",
    "for_each_netdev",
    "get_net_ns_by_inode",
    "get_net_ns_by_fd",
    "netdev_for_each_tx_queue",
    "netdev_get_by_index",
    "netdev_get_by_name",
    "netdev_name",
    "netdev_priv",
    "sk_fullsock",
    "sk_nulls_for_each",
    "skb_shinfo",
    "is_pp_page",
    "netdev_ipv4_addrs",
    "netdev_ipv6_addrs",
    "neigh_table_for_each_neighbor",
)


_S_IFMT = 0o170000
_S_IFSOCK = 0o140000


def SOCKET_I(inode: Object) -> Object:
    """
    Get a socket from an inode referring to the socket.

    :param inode: ``struct inode *``
    :return: ``struct socket *``
    :raises ValueError: If *inode* does not refer to a socket
    """
    if inode.i_mode & _S_IFMT != _S_IFSOCK:
        raise ValueError("not a socket inode")

    return container_of(inode, "struct socket_alloc", "vfs_inode").socket.address_of_()


def SOCK_INODE(sock: Object) -> Object:
    """
    Get the inode of a socket.

    :param sock: ``struct socket *``
    :return: ``struct inode *``
    """
    return container_of(sock, "struct socket_alloc", "socket").vfs_inode.address_of_()


@takes_program_or_default
def for_each_net(prog: Program) -> Iterator[Object]:
    """
    Iterate over all network namespaces in the system.

    :return: Iterator of ``struct net *`` objects.
    """
    return list_for_each_entry(
        "struct net", prog["net_namespace_list"].address_of_(), "list"
    )


_CLONE_NEWNET = 0x40000000


def get_net_ns_by_inode(inode: Object) -> Object:
    """
    Get a network namespace from a network namespace NSFS inode, e.g.
    ``/proc/$PID/ns/net`` or ``/var/run/netns/$NAME``.

    :param inode: ``struct inode *``
    :return: ``struct net *``
    :raises ValueError: if *inode* is not a network namespace inode
    """
    if inode.i_fop != inode.prog_["ns_file_operations"].address_of_():
        raise ValueError("not a namespace inode")

    ns = cast("struct ns_common *", inode.i_private)
    # Linux kernel commit 4055526d3574 ("ns: move ns type into struct
    # ns_common") (in v6.18) moved the type from struct proc_ns_operations to
    # struct ns_common.
    try:
        ns_type = ns.ns_type
    except AttributeError:
        ns_type = ns.ops.type
    if ns_type != _CLONE_NEWNET:
        raise ValueError("not a network namespace inode")

    return container_of(ns, "struct net", "ns")


def get_net_ns_by_fd(task: Object, fd: IntegerLike) -> Object:
    """
    Get a network namespace from a task and a file descriptor referring to a
    network namespace NSFS inode, e.g. ``/proc/$PID/ns/net`` or
    ``/var/run/netns/$NAME``.

    :param task: ``struct task_struct *``
    :param fd: File descriptor.
    :return: ``struct net *``
    :raises ValueError: If *fd* does not refer to a network namespace inode
    """
    return get_net_ns_by_inode(fget(task, fd).f_inode)


def netdev_for_each_tx_queue(dev: Object) -> Iterator[Object]:
    """
    Iterate over all TX queues for a network device.

    :param dev: ``struct net_device *``
    :return: Iterator of ``struct netdev_queue *`` objects.
    """
    for i in range(dev.num_tx_queues):
        yield dev._tx + i


@takes_object_or_program_or_default
def for_each_netdev(prog: Program, net: Optional[Object]) -> Iterator[Object]:
    """
    Iterate over all network devices in a namespace.

    :param net: ``struct net *``. Defaults to the initial network namespace if
        given a :class:`~drgn.Program` or :ref:`omitted <default-program>`.
    :return: Iterator of ``struct net_device *`` objects.
    """
    if net is None:
        net = prog["init_net"]
    return list_for_each_entry(
        "struct net_device", net.dev_base_head.address_of_(), "dev_list"
    )


def netdev_name(dev: Object) -> bytes:
    """
    Get the name of a network device.

    >>> netdev_name(dev)
    b'lo'

    :param dev: ``struct net_device *``
    """
    return dev.name.string_()


_NETDEV_HASHBITS = 8
_NETDEV_HASHENTRIES = 1 << _NETDEV_HASHBITS


@takes_object_or_program_or_default
def netdev_get_by_index(
    prog: Program, net: Optional[Object], ifindex: IntegerLike
) -> Object:
    """
    Get the network device with the given interface index number.

    :param net: ``struct net *``. Defaults to the initial network namespace if
        given a :class:`~drgn.Program` or :ref:`omitted <default-program>`.
    :param ifindex: Network interface index number.
    :return: ``struct net_device *`` (``NULL`` if not found)
    """
    if net is None:
        net = prog["init_net"]
    ifindex = operator.index(ifindex)

    head = net.dev_index_head[ifindex & (_NETDEV_HASHENTRIES - 1)]
    for netdev in hlist_for_each_entry("struct net_device", head, "index_hlist"):
        if netdev.ifindex.value_() == ifindex:
            return netdev

    return NULL(prog, "struct net_device *")


@takes_object_or_program_or_default
def netdev_get_by_name(
    prog: Program, net: Optional[Object], name: Union[str, bytes]
) -> Object:
    """
    Get the network device with the given interface name.

    :param net: ``struct net *``. Defaults to the initial network namespace if
        given a :class:`~drgn.Program` or :ref:`omitted <default-program>`.
    :param name: Network interface name.
    :return: ``struct net_device *`` (``NULL`` if not found)
    """
    if net is None:
        net = prog["init_net"]
    if isinstance(name, str):
        name = name.encode()

    for dev in for_each_netdev(net):
        if netdev_name(dev) == name:
            return dev
    return NULL(prog, "struct net_device *")


def netdev_priv(dev: Object, type: Union[str, Type] = "void") -> Object:
    """
    Return the private data of a network device.

    >>> dev = netdev_get_by_name("wlp0s20f3")
    >>> netdev_priv(dev)
    (void *)0xffff9419c9dec9c0
    >>> netdev_priv(dev, "struct ieee80211_sub_if_data")
    *(struct ieee80211_sub_if_data *)0xffff9419c9dec9c0 = {
        ...
    }

    :param dev: ``struct net_device *``
    :param type: Type of private data.
    :return: ``type *``
    """
    prog = dev.prog_
    try:
        offset = prog.cache["net_device_aligned_size"]
    except KeyError:
        # 31 is NETDEV_ALIGN - 1
        offset = (sizeof(prog.type("struct net_device")) + 31) & ~31
        prog.cache["net_device_aligned_size"] = offset
    return Object(prog, prog.pointer_type(prog.type(type)), dev.value_() + offset)


def sk_fullsock(sk: Object) -> bool:
    """
    Check whether a socket is a full socket, i.e., not a time-wait or request
    socket.

    :param sk: ``struct sock *``
    """
    prog = sk.prog_
    state = sk.__sk_common.skc_state.value_()
    return state != prog["TCP_SYN_RECV"] and state != prog["TCP_TIME_WAIT"]


def sk_nulls_for_each(head: Object) -> Iterator[Object]:
    """
    Iterate over all the entries in a nulls hash list of sockets specified by
    ``struct hlist_nulls_head`` head.

    :param head: ``struct hlist_nulls_head *``
    :return: Iterator of ``struct sock *`` objects.
    """
    return hlist_nulls_for_each_entry("struct sock", head, "__sk_common.skc_nulls_node")


def skb_shinfo(skb: Object) -> Object:
    """
    Get the shared info for a socket buffer.

    :param skb: ``struct sk_buff *``
    :return: ``struct skb_shared_info *``
    """
    prog = skb.prog_
    try:
        NET_SKBUFF_DATA_USES_OFFSET = prog.cache["NET_SKBUFF_DATA_USES_OFFSET"]
    except KeyError:
        NET_SKBUFF_DATA_USES_OFFSET = sizeof(prog.type("long")) > 4
        prog.cache["NET_SKBUFF_DATA_USES_OFFSET"] = NET_SKBUFF_DATA_USES_OFFSET
    if NET_SKBUFF_DATA_USES_OFFSET:
        return cast("struct skb_shared_info *", skb.head + skb.end)
    else:
        return cast("struct skb_shared_info *", skb.end)


def _poison_pointer_delta(prog: Program) -> int:
    # The value of POISON_POINTER_DELTA depends on
    # CONFIG_ILLEGAL_POINTER_VALUE, which varies by architecture and kernel
    # version. To avoid hard-coding values, derive the value from
    # TIMER_ENTRY_STATIC, which we can find in any statically-defined timer.
    # This still requires hard-coding an offset, but that offset is the same on
    # all architectures and kernel versions since Linux kernel commit
    # b8a0255db958 ("include/linux/poison.h: use POISON_POINTER_DELTA for
    # poison pointers") (in v4.5).
    try:
        return prog.cache["POISON_POINTER_DELTA"]
    except KeyError:
        pass
    TIMER_ENTRY_STATIC = prog["poll_spurious_irq_timer"].entry.next.value_()
    POISON_POINTER_DELTA = TIMER_ENTRY_STATIC - 0x300
    prog.cache["POISON_POINTER_DELTA"] = POISON_POINTER_DELTA
    return POISON_POINTER_DELTA


def is_pp_page(page: Object) -> bool:
    """
    Check if given page is a page_pool page.

    :param page: ``struct page *``
    :raises NotImplementedError: If page_pool pages cannot be identified on
        this kernel. This is the case from Linux 4.18 (when page_pool was
        introduced) up to and including Linux 5.13.
    """
    prog = page.prog_

    # Since Linux kernel commit db359fccf212 ("mm: introduce a new page type
    # for page pool in page type") (in v7.1), page pool pages are identified
    # by PGTY_netpp in page->page_type instead of the pp_magic signature.
    try:
        PGTY_netpp = prog["PGTY_netpp"]
    except KeyError:
        pass
    else:
        return page.page_type.value_() >> 24 == PGTY_netpp.value_()

    PP_SIGNATURE = _poison_pointer_delta(prog) + 0x40
    PP_MAGIC_MASK = ~0x3

    try:
        return (page.pp_magic & PP_MAGIC_MASK) == PP_SIGNATURE
    except AttributeError:
        pass
    # Before Linux kernel commit ff7d6b27f894 ("page_pool: refurbish version of
    # page_pool code") (in v4.18), page_pool didn't exist.
    try:
        prog.type("struct page_pool")
    except LookupError:
        return False
    # Between that and Linux kernel commit c07aea3ef4d4 ("mm: add a signature
    # in struct page") (in v5.14), there is no way to identify page_pool pages.
    raise NotImplementedError("page_pool pages cannot be identified before Linux 5.14")


def netdev_ipv4_addrs(dev: Object) -> List[ipaddress.IPv4Address]:
    """
    Get the list of IPV4 addresses associated with a network device.

    :param dev: ``struct net_device *``
    """
    ips = []
    prog = dev.prog_
    ip_ptr = dev.ip_ptr.read_()
    if ip_ptr:
        ifa = ip_ptr.ifa_list.read_()
        while ifa:
            addr_bytes = prog.read(
                ifa.ifa_address.address_,  # type: ignore[arg-type]  # address can't be None.
                4,
            )
            ips.append(ipaddress.IPv4Address(addr_bytes))
            ifa = ifa.ifa_next.read_()
    return ips


def netdev_ipv6_addrs(dev: Object) -> List[ipaddress.IPv6Address]:
    """
    Get the list of IPV6 addresses associated with a network device.

    :param dev: ``struct net_device *``
    """
    ips = []
    prog = dev.prog_
    ip_ptr = dev.ip6_ptr.read_()
    if ip_ptr:
        for addr in list_for_each_entry(
            "struct inet6_ifaddr",
            dev.ip6_ptr.addr_list.address_of_(),
            "if_list",
        ):
            addr_bytes = prog.read(
                addr.addr.in6_u.u6_addr8.address_,  # type: ignore[arg-type]  # address can't be None.
                16,
            )
            ips.append(ipaddress.IPv6Address(addr_bytes))
    return ips


def neigh_table_for_each_neighbor(table: Object) -> Iterator[Object]:
    """
    Get the neighbour object from the neighbour hash table

    :param dev: ``struct neigh_hash_table *``
    """
    n_buckets = 1 << table.hash_shift.value_()
    # Since Linux kernel commit 41b3caa7c076 ("neighbour: Add hlist_node to
    # struct neighbour") (in v6.13), neighbors are in an hlist in struct
    # neigh_hash_table::hash_heads. Before that, they are in a manual linked
    # list in struct neigh_hash_table::hash_buckets.
    try:
        hash_heads = table.hash_heads
    except AttributeError:
        for neigh in table.hash_buckets[:n_buckets]:
            while neigh := neigh.read_():
                yield neigh
                neigh = neigh.next
    else:
        for head in hash_heads[:n_buckets]:
            yield from hlist_for_each_entry(
                "struct neighbour", head.address_of_(), "hash"
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re
import textwrap
from typing import Any, List

from _drgn_util.typingutils import copy_func_params


def multiline_wrap(text: str, width: int, *, indent: str = "") -> List[str]:
    """
    Wrap text containing multiple paragraphs or line blocks, returning lines.

    This dedents *text* and splits it into paragraphs, which are wrapped, and
    line blocks, which are not wrapped and have their line breaks preserved.

    Paragraphs are groups of lines separated by one or more blank lines.

    Line blocks are groups of lines starting with the '|' character.

    >>> s = '''
    ... This is a paragraph, which will
    ... be wrapped.
    ...
    ... |This is a line block.
    ... |It will not be wrapped.
    ... '''
    ...
    >>> multiline_wrap(s, width=80)
    ['This is a paragraph, which will be wrapped.', '', 'This is a line block.', 'It will not be wrapped.']

    :param text: Text to wrap.
    :param width: Maximum width of lines.
    :param indent: String to prepend to each line.
    :return: List of lines (without newlines).
    """
    lines = []
    blank = False
    for match in re.finditer(
        r"""
        (?P<paragraph>
            # A paragraph is one or more lines not starting with '|' and
            # containing at least one non-whitespace character.
            (?:
                (?:
                    # Each line must start with either a non-whitespace,
                    # non-'|' character...
                    [^\s|]
                    |
                    # ... or any amount of non-newline whitespace followed by a
                    # non-whitespace character.
                    [^\S\n]+\S
                ).*(?:\n|$)
            )+
        )
        |
        (?P<lineblock>
            # A line block is one or more lines starting with '|'.
            (?:\|.*(?:\n|$))+
        )
        |
        (?:
            # Ignore one or more lines consisting of only whitespace.
            \s*(?:\n|$)
        )
        """,
        textwrap.dedent(text),
        flags=re.X,
    ):
        lastgroup = match.lastgroup
        if lastgroup == "lineblock":
            if blank:
                lines.append("")
                blank = False
            lines.extend(
                re.sub(r"^\|", "", match.group(lastgroup), flags=re.M).splitlines()
            )
        elif lastgroup == "paragraph":
            if blank:
                lines.append("")
                blank = False
            lines.extend(
                textwrap.wrap(
                    match.group(lastgroup),
                    width,
                    initial_indent=indent,
                    subsequent_indent=indent,
                )
            )
        elif lines:
            blank = True
    return lines


@copy_func_params(multiline_wrap)
def multiline_fill(*args: Any, **kwargs: Any) -> str:
    r"""
    Wrap text containing multiple paragraphs or line blocks, returning a
    string.

    This is equivalent to

    .. code-block:: python3

        "\n".join(multiline_wrap(text, ...))

    >>> s = '''
    ... This is a paragraph, which will
    ... be wrapped.
    ...
    ... |This is a line block.
    ... |It will not be wrapped.
    ... '''
    ...
    >>> print(multiline_fill(s, width=80))
    This is a paragraph, which will be wrapped.

    This is a line block.
    It will not be wrapped.

    :return: Wrapped string (without trailing newline).
    """
    return "\n".join(multiline_wrap(*args, **kwargs))



# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
import argparse
from typing import Any, Iterable, Literal, Optional

from drgn import Architecture, FaultError, MainModule, Object, Program, StackFrame
from drgn.commands import argument, drgn_argument, mutually_exclusive_group
from drgn.commands._crash._rd import _print_memory
from drgn.commands._crash.common import (
    Cpuspec,
    CrashDrgnCodeBuilder,
    _crash_foreach_subcommand,
    _TaskSelector,
    crash_command,
    parse_cpuspec,
    print_task_header,
)
from drgn.helpers.common.stack import print_registers
from drgn.helpers.linux.sched import cpu_curr
from drgn.helpers.linux.stack import LinuxKernelStack, StackKind, kernel_stack_trace


def _sp_align(trace: LinuxKernelStack) -> int:
    # Some architectures may have shorter address widths than their register
    # width. We want to align the stack frames nicely, and not waste space on
    # extra zeros. Determine the width for the stack pointer field empirically
    # rather than setting it to 8 or 16 based on IS_64_BIT. Subtract 2 to
    # account for the "0x" prefix. Swallow the LookupError in case the sp is not
    # known.
    width = 2
    for segment in trace.segments:
        for frame in segment.frames:
            try:
                width = max(width, len(hex(frame.sp)))
            except LookupError:
                pass
    return width - 2


def _print_frame(
    prog: Program,
    index: int,
    sp: Optional[int],
    frame: StackFrame,
    sp_width: int,
    drgn_style: bool = False,
    show_file_line: bool = False,
) -> None:
    index_str = f"#{index}"

    if drgn_style:
        try:
            source_info = " ({}:{}:{})".format(*frame.source())
        except LookupError:
            source_info = ""
        print(f"{index_str:>3s} {frame.name}{source_info}")
    else:
        mod_text = ""
        try:
            module = prog.module(frame.pc - int(frame.pc > 0 and not frame.interrupted))
            if not isinstance(module, MainModule):
                mod_text = f" [{module.name}]"
        except LookupError:
            pass
        if frame.is_inline:
            print(
                f"{index_str:>3s} {'(inline)':{sp_width + 2}s} {frame.name}{mod_text}"
            )
        elif sp is None:
            print(
                f"{index_str:>3s} [{'?' * sp_width}] {frame.name} at {frame.pc:x}{mod_text}"
            )
        else:
            print(
                f"{index_str:>3s} [{frame.sp:0{sp_width}x}] {frame.name} at {frame.pc:x}{mod_text}"
            )
        if show_file_line:
            try:
                file, line, _ = frame.source()
                print(f"    {file}: {line}")
            except LookupError:
                pass


def _print_variables(prog: Program, frame: StackFrame) -> None:
    for name in frame.locals():
        value = frame[name]
        if not value.absent_:
            try:
                val_str = value.format_(dereference=False).replace("\n", "\n    ")
                print(f"    {name} = {val_str}")
            except FaultError:
                pass


def _maybe_print_mem(
    prog: Program,
    prev_sp: Optional[int],
    sp: Optional[int],
    annotate: Literal[None, "symbols", "slab", "verbose"],
) -> None:
    if prev_sp is None or sp is None:
        return
    # The stack grows down on all supported architectures, and we should never
    # see a gap between frames larger than THREAD_SIZE.
    if sp <= prev_sp or sp - prev_sp >= prog["THREAD_SIZE"]:
        return

    # _print_memory() already asserts this, so let's just yell at the type
    # checker to pervent it complaining.
    unit: Literal[1, 2, 4, 8] = prog.address_size()  # type: ignore

    _print_memory(
        prog,
        prev_sp,
        prog.read(prev_sp, sp - prev_sp),
        unit=unit,
        show_ascii=False,
        annotate=annotate,
        indent="    ",
        address_pad="0",
    )


def _print_one_task(
    trace: LinuxKernelStack,
    show_file_line: bool = False,
    drgn_style: bool = False,
    show_variables: bool = False,
    show_mem: bool = False,
    mem_annotate: Literal[None, "symbols", "slab", "verbose"] = None,
) -> None:
    prog = trace.prog
    sp_width = _sp_align(trace)

    i = -1

    # On s390x, the stack pointer points to the top (high address) of a 160-byte
    # stack frame record. However, in practice the kernel is built with
    # -mpacked-stack, so the frames are not actually 160 bytes: only the bottom
    # few words (i.e. the low addresses) of the record are present. The rest of
    # the record is not present. This means that in practice, each frame's stack
    # pointer is pointing several words "too high", pointing into their callers'
    # stack frames. To correct for this, we have to offset the stack pointers by
    # 160 bytes too.
    sp_offset = 0
    assert prog.platform is not None
    if prog.platform.arch == Architecture.S390X:
        sp_offset = 160

    prev_sp = None
    for segment_idx, segment in enumerate(trace.segments):
        if segment_idx != 0:
            print_registers(prog, segment.frames[0].registers())

        # Don't show the last userspace stack segment
        if segment_idx == len(trace.segments) - 1 and segment.kind == StackKind.USER:
            break

        for i, frame in enumerate(segment.frames, i + 1):
            try:
                sp = frame.sp + sp_offset
            except LookupError:
                sp = None
            if show_mem:
                _maybe_print_mem(prog, prev_sp, sp, mem_annotate)
            _print_frame(prog, i, sp, frame, sp_width, drgn_style, show_file_line)
            if show_variables:
                _print_variables(prog, frame)
            prev_sp = sp or prev_sp

        # Print separators for exception/interrupt stacks, like crash does
        if segment.kind not in (StackKind.USER, StackKind.TASK, StackKind.UNKNOWN):
            print(f"--- <{segment.kind.name} stack> ---")


def _print_bt(
    prog: Program, tasks: Iterable[Object], args: argparse.Namespace
) -> Optional[LinuxKernelStack]:
    trace = None
    for i, task in enumerate(tasks):
        if i > 0:
            print()
        print_task_header(task)
        try:
            trace = kernel_stack_trace(task)
        except ValueError as e:
            if "running task" in str(e):
                print("(active)")
                continue
            else:
                raise
        _print_one_task(
            trace,
            drgn_style=args.drgn_style,
            show_file_line=args.show_file_line,
            show_variables=args.show_variables,
            show_mem=args.frame,
            mem_annotate=args.annotate,
        )
    return trace


def _append_bt_code(code: CrashDrgnCodeBuilder, args: argparse.Namespace) -> None:
    """Generate the code to unwind a single task according to the arguments"""
    code.add_from_import("drgn.helpers.linux.stack", "kernel_stack_trace")
    code.add_from_import("drgn.helpers.common.format", "escape_ascii_string")

    if not code.has_declaration("task"):
        code.add_from_import("drgn.helpers.linux.sched", "cpu_curr")
        code.append("task = cpu_curr(prog, cpu)\n")
    elif not code.has_declaration("cpu"):
        code.add_from_import("drgn.helpers.linux.sched", "task_cpu")
        code.append("cpu = task_cpu(task)\n")

    code.append(
        """\
trace = kernel_stack_trace(task)
pid = task.pid.value_()
comm = escape_ascii_string(task.comm.string_())
"""
    )
    if args.show_file_line or args.show_variables:
        code.append(
            """\
for segment in trace.segments:
    registers = segment.frames[0].registers()
    segment_kind = segment.kind
    for frame in segment.frames:
"""
        )
    if args.show_file_line:
        code.append(
            """\
        try:
            file, line, col = frame.source()
        except LookupError:
            pass
"""
        )
    if args.show_variables:
        code.append(
            """\
        for var in frame.locals():
            value = frame[var]
"""
        )


@_crash_foreach_subcommand(
    arguments=(
        mutually_exclusive_group(
            argument(
                "-f",
                dest="frame",
                action="store_true",
                help="display all stack data contained in a frame",
            ),
            argument(
                "-F",
                dest="frame_symbolic",
                action="count",
                default=0,
                help="similar to -f, except that the stack data is displayed "
                "symbolically where appropriate; if the stack data references a "
                "slab cache object, the name of the slab cache will be displayed "
                "in brackets. If -F is entered twice, and the stack data references "
                "a slab cache object, then both the address and the name of the "
                "slab cache will be displayed in brackets",
            ),
        ),
        argument(
            "-l",
            dest="show_file_line",
            action="store_true",
            help="show file and line number of each stack trace text location",
        ),
        argument(
            "-d",
            dest="drgn_style",
            action="store_true",
            help="format the stack frames in drgn's style rather than crash's",
        ),
        argument(
            "-V",
            dest="show_variables",
            action="store_true",
            help="print any local variable values which drgn can determine",
        ),
        drgn_argument,
    ),
)
def _crash_foreach_bt(
    task_selector: _TaskSelector, args: argparse.Namespace
) -> Optional[LinuxKernelStack]:
    prog = task_selector.prog

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            _append_bt_code(code, args)
        code.print()
        return None

    args.annotate = None
    if args.frame_symbolic >= 1:
        args.frame = True
        args.annotate = "slab"
    if args.frame_symbolic >= 2:
        args.annotate = "verbose"

    return _print_bt(prog, task_selector.tasks(), args)


@crash_command(
    description="print stack traces",
    long_description="""
    Print the kernel stack trace for a given PID or task

    If no PID or task is given, print the stack trace for the current context
    task.
    """,
    arguments=(
        mutually_exclusive_group(
            # Work around https://github.com/python/cpython/issues/72795
            # before Python 3.13.
            argument(
                "tasks",
                metavar="pid|task",
                type="pid_or_task",
                nargs="*",
                default=[],
                help="print stack trace for task(s), given as a decimal process ID or"
                " hexadecimal ``task_struct`` address.",
            ),
            argument(
                "-c",
                dest="cpus",
                type=str,
                help="print stack trace for one or more CPUs, which can be specified"
                " using the format '3', '1,8,9', '1-23', or '1,8,9-14'.",
            ),
            argument(
                "-a",
                dest="all_cpus",
                action="store_true",
                help="print stack trace for all CPUs",
            ),
            argument(
                "-p",
                dest="panic",
                action="store_true",
                help="print stack trace for the panic task",
            ),
        ),
        mutually_exclusive_group(
            argument(
                "-f",
                dest="frame",
                action="store_true",
                help="display all stack data contained in a frame",
            ),
            argument(
                "-F",
                dest="frame_symbolic",
                action="count",
                default=0,
                help="similar to -f, except that the stack data is displayed "
                "symbolically where appropriate; if the stack data references a "
                "slab cache object, the name of the slab cache will be displayed "
                "in brackets. If -F is entered twice, and the stack data references "
                "a slab cache object, then both the address and the name of the "
                "slab cache will be displayed in brackets",
            ),
        ),
        argument(
            "-l",
            dest="show_file_line",
            action="store_true",
            help="show file and line number of each stack trace text location",
        ),
        argument(
            "-d",
            dest="drgn_style",
            action="store_true",
            help="format the stack frames in drgn's style rather than crash's",
        ),
        argument(
            "-V",
            dest="show_variables",
            action="store_true",
            help="print any local variable values which drgn can determine",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_bt(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> Optional[LinuxKernelStack]:
    # The bt command supports use via foreach, and thus it should support the
    # full range of task selectors. However, it also supports selection via
    # CPUs. Rather than try to shoehorn this functionality into _TaskSelector,
    # we handle CPU selection separately. Most of the code still ends up shared.
    if args.tasks:
        return _crash_foreach_bt(_TaskSelector(prog, args.tasks), args)
    elif args.all_cpus:
        cpuspec = Cpuspec(all=True)
    elif args.cpus:
        cpuspec = parse_cpuspec(args.cpus)
    elif args.panic:
        cpuspec = Cpuspec(panic=True)
    else:
        return _crash_foreach_bt(_TaskSelector(prog, [None]), args)

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        with code.begin_cpuspec_loop(cpuspec):
            _append_bt_code(code, args)
        code.print()
        return None

    args.annotate = None
    if args.frame_symbolic >= 1:
        args.frame = True
        args.annotate = "slab"
    if args.frame_symbolic >= 2:
        args.annotate = "verbose"

    tasks = (cpu_curr(prog, cpu) for cpu in cpuspec.cpus(prog))
    return _print_bt(prog, tasks, args)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Interprocess Communication
--------------------------

The ``drgn.helpers.linux.ipc`` module provides helpers for working with System
V interprocess communication mechanisms.
"""

import operator
from typing import Iterator, Optional

from drgn import NULL, IntegerLike, Object, ObjectNotFoundError, Program, container_of
from drgn.helpers.common.format import decode_flags
from drgn.helpers.common.prog import (
    takes_object_or_program_or_default,
    takes_program_or_default,
)
from drgn.helpers.linux.idr import idr_find, idr_for_each

__all__ = (
    "decode_sysv_shm_flags",
    "decode_sysv_shm_mode_flags",
    "find_sysv_msg_queue",
    "find_sysv_sem_array",
    "find_sysv_shm",
    "for_each_sysv_msg_queue",
    "for_each_sysv_sem_array",
    "for_each_sysv_shm",
)

_IPC_SEM_IDS = 0
_IPC_MSG_IDS = 1
_IPC_SHM_IDS = 2


def _for_each_sysv_ipc(
    prog: Program, ns: Optional[Object], kind: int, type_name: str, member: str
) -> Iterator[Object]:
    if ns is None:
        ns = prog["init_ipc_ns"]
    for _, entry in idr_for_each(ns.ids[kind].ipcs_idr.address_of_()):
        yield container_of(entry, type_name, member)


def _find_sysv_ipc(
    prog: Program,
    ns: Optional[Object],
    id: IntegerLike,
    kind: int,
    type_name: str,
    member: str,
) -> Object:
    if ns is None:
        ns = prog["init_ipc_ns"]
    try:
        ipcmni_seq_shift = prog["ipc_mni_shift"].value_()
    except ObjectNotFoundError:
        ipcmni_seq_shift = 15
    id = operator.index(id)

    entry = idr_find(
        ns.ids[kind].ipcs_idr.address_of_(), id & ((1 << ipcmni_seq_shift) - 1)
    )
    if not entry:
        return NULL(prog, prog.pointer_type(prog.type(type_name)))
    entry = container_of(entry, type_name, member)
    if entry.member_(member).seq.value_() != (id >> ipcmni_seq_shift):
        return NULL(prog, entry.type_)
    return entry


@takes_object_or_program_or_default
def for_each_sysv_msg_queue(prog: Program, ns: Optional[Object]) -> Iterator[Object]:
    """
    Iterate over all System V message queues in a namespace.

    :param ns: ``struct ipc_namespace *``. Defaults to the initial IPC
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :return: Iterator of ``struct msg_queue *`` objects.
    """
    return _for_each_sysv_ipc(prog, ns, _IPC_MSG_IDS, "struct msg_queue", "q_perm")


@takes_object_or_program_or_default
def find_sysv_msg_queue(prog: Program, ns: Optional[Object], id: IntegerLike) -> Object:
    """
    Find a System V message queue by ID.

    :param ns: ``struct ipc_namespace *``. Defaults to the initial IPC
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :param id: Message queue identifier.
    :return: ``struct msg_queue *`` (``NULL`` if not found)
    """
    return _find_sysv_ipc(prog, ns, id, _IPC_MSG_IDS, "struct msg_queue", "q_perm")


@takes_object_or_program_or_default
def for_each_sysv_shm(prog: Program, ns: Optional[Object]) -> Iterator[Object]:
    """
    Iterate over all System V shared memory segments in a namespace.

    :param ns: ``struct ipc_namespace *``. Defaults to the initial IPC
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :return: Iterator of ``struct shmid_kernel *`` objects.
    """
    return _for_each_sysv_ipc(prog, ns, _IPC_SHM_IDS, "struct shmid_kernel", "shm_perm")


@takes_object_or_program_or_default
def find_sysv_shm(prog: Program, ns: Optional[Object], id: IntegerLike) -> Object:
    """
    Find a System V shared memory segment by ID.

    :param ns: ``struct ipc_namespace *``. Defaults to the initial IPC
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :param id: Shared memory segment identifier.
    :return: ``struct shmid_kernel *`` (``NULL`` if not found)
    """
    return _find_sysv_ipc(prog, ns, id, _IPC_SHM_IDS, "struct shmid_kernel", "shm_perm")


def decode_sysv_shm_flags(shm: Object) -> str:
    """
    Get a human-readable representation of the flags set on a System V shared
    memory segment.

    >>> decode_sysv_shm_flags(shm)
    'SHM_LOCKED'

    :param shm: ``struct shmid_kernel *``
    """
    return decode_sysv_shm_mode_flags(shm.shm_perm.mode)


@takes_program_or_default
def decode_sysv_shm_mode_flags(prog: Program, value: IntegerLike) -> str:
    """
    Get a human-readable representation of the flags in the ``mode`` value of a
    System V shared memory segment.

    >>> decode_sysv_shm_mode_flags(0o2777)
    'SHM_LOCKED'

    :param mode: ``umode_t``
    """
    return decode_flags(
        operator.index(value) & ~0o777,
        # These are ABI, so we don't have to worry about them changing.
        (("SHM_DEST", 0o1000), ("SHM_LOCKED", 0o2000)),
        bit_numbers=False,
    )


@takes_object_or_program_or_default
def for_each_sysv_sem_array(prog: Program, ns: Optional[Object]) -> Iterator[Object]:
    """
    Iterate over all System V semaphore arrays in a namespace.

    :param ns: ``struct ipc_namespace *``. Defaults to the initial IPC
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :return: Iterator of ``struct sem_array *`` objects.
    """
    return _for_each_sysv_ipc(prog, ns, _IPC_SEM_IDS, "struct sem_array", "sem_perm")


@takes_object_or_program_or_default
def find_sysv_sem_array(prog: Program, ns: Optional[Object], id: IntegerLike) -> Object:
    """
    Find a System V semaphore array by ID.

    :param ns: ``struct ipc_namespace *``. Defaults to the initial IPC
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :param id: Semaphore array identifier.
    :return: ``struct sem_array *`` (``NULL`` if not found)
    """
    return _find_sysv_ipc(prog, ns, id, _IPC_SEM_IDS, "struct sem_array", "sem_perm")



# Copyright (c) 2025, Oracle and/or its affiliates.
"""
Tools for creating stack traces of userspace tasks, from the kernel

This script contains tools that enable creating stack traces for userspace
tasks, when debugging a kernel program. This requires having access to the
userspace pages, which is not common for core dumps (see makedumpfile's -d
option), however it is available for /proc/kcore and /proc/vmcore. The script
contains two approaches:

1. By directly creating a Program to represent a process, which reads memory via
   the kernel Program. This allows directly printing stack traces. This approach
   works well with /proc/kcore, but in the kexec/kdump environment it may
   require too much memory, as well as access to the full root filesystem.
2. By dumping userspace stack memory and some metadata. Later, a second call can
   read this information and actually create the stack trace. This approach
   works better in a kexec environment, because the root filesystem is not
   required, and userspace debuginfo is not consulted.

This script is tested on x86_64 and aarch64, and it's especially suited for
Fedora and its derivatives, due to their inclusion of ".eh_frame" on runtime
binaries, as well as ".gnu_debugdata" sections for address to symbol resolution.
"""
import argparse
import base64
import ctypes.util
import fnmatch
import gzip
import json
import logging
import os
import struct
import sys
import warnings
from bisect import bisect_left
from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import drgn
from drgn import Architecture
from drgn import FaultError
from drgn import Object
from drgn import Program
from drgn import ProgramFlags
from drgn import sizeof
from drgn import StackTrace
from drgn.helpers.linux import access_remote_vm
from drgn.helpers.linux import cpu_curr
from drgn.helpers.linux import d_path
from drgn.helpers.linux import find_task
from drgn.helpers.linux import for_each_online_cpu
from drgn.helpers.linux import for_each_task
from drgn.helpers.linux import for_each_task_in_group
from drgn.helpers.linux import for_each_vma
from drgn.helpers.linux import task_cpu
from drgn.helpers.linux import task_on_cpu
from drgn.helpers.linux import task_state_to_char
from drgn.helpers.linux import vma_find


log = logging.getLogger("drgn.pstack")


class CommaList(argparse.Action):
    """
    Action that allows specifying an option multiple times, with comma-separated
    values
    """

    def __init__(self, *args, element_type=str, **kwargs) -> None:
        self.element_type = element_type
        return super().__init__(*args, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        value: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        assert isinstance(value, str)
        result = getattr(namespace, self.dest, []) or []
        for element in value.split(","):
            result.append(self.element_type(element))
        setattr(namespace, self.dest, result)


def task_saved_pt_regs(task: Object) -> Object:
    """
    Return the userspace registers for the given task struct

    This returns the registers which were saved on entry to the kernel. For
    vmcores generated via kexec and /proc/vmcore, all userspace tasks will have
    registers stored on the stack, because every CPU should be interrupted and
    halted. However, for vmcores which were created by a hypervisor, or for
    live systems, userspace tasks may be directly executing, and any data stored
    on the kernel stack is stale. Drgn does not provide an easy API to get this
    info, but you can tell based on whether the stack pointer is a user or
    kernel address.

    :param task: the ``struct task_struct *`` of this task
    :returns: a ``struct pt_regs`` value object
    """
    prog = task.prog_
    # The pt_regs is dumped at the top of the stack. The stack size may vary,
    # but it gets a guard page on top, and there's sometimes padding. See
    # TOP_OF_STACK_PADDING in arch/x86/include/asm/thread_info.h -- for x86_64,
    # if FRED is enabled, then there is 16 bytes of padding, otherwise 0.
    # an offset of 16 bytes for 64-bit.
    try:
        prog.symbol("fred_rsp0")
        padding = 16
    except LookupError:
        padding = 0
    regs_addr = (
        task.stack_vm_area.addr.value_()
        + task.stack_vm_area.size.value_()
        - sizeof(prog.type("struct pt_regs"))
        - prog["PAGE_SIZE"]
        - padding
    )
    return Object(prog, "struct pt_regs", address=regs_addr)


def task_running_pt_regs(kstack: StackTrace) -> Object:
    """
    Create a ``struct pt_regs`` object from the top frame of a stack trace

    This returns the registers for a task that is/was actively running. They
    should be stored in the core dump metadata (e.g. PRSTATUS), and we can get
    at them via drgn's stack trace object. Drgn's kernel Program won't be able
    to unwind it anyway.

    :param kstack: The kernel stack trace.
    :returns: A ``struct pt_regs`` value object containing the user-space registers.
    """
    prog = kstack.prog
    pt_regs = {}
    tp = prog.type("struct pt_regs")
    if prog.platform.arch == Architecture.X86_64:
        rename = {
            "rip": "ip",
            "rbp": "bp",
            "rax": "ax",
            "rbx": "bx",
            "rcx": "cx",
            "rdx": "dx",
            "rdi": "di",
            "rsi": "si",
            "rsp": "sp",
            "rflags": "flags",
        }
        for name, value in kstack[0].registers().items():
            if name in rename:
                name = rename[name]
            try:
                tp.member(name)
            except LookupError:
                continue
            pt_regs[name] = value
    elif prog.platform.arch == Architecture.AARCH64:
        pt_regs["regs"] = [0] * 31
        pt_regs["pc"] = kstack[0].pc
        for name, value in kstack[0].registers().items():
            if name[0] == "x":
                pt_regs["regs"][int(name[1:])] = value
            elif name == "lr":  # an alias for x30
                pt_regs["regs"][30] = value
            else:
                try:
                    tp.member(name)
                    pt_regs[name] = value
                except LookupError:
                    pass
    else:
        raise NotImplementedError(
            f"Support for {prog.platform.arch} is not implemented"
        )

    return Object(prog, "struct pt_regs", value=pt_regs)


def make_fake_pt_regs(up: Program, data: bytes) -> Object:
    """
    Create a fake ``struct pt_regs`` to convince drgn to unwind a thread

    Drgn's unwinder will accept any object that looks like a ``struct pt_regs``
    (a correctly-named struct of the correct size) and use it as the initial
    registers for a stack unwind. This function can take the bytes of a real
    pt_regs object, and a Program, and return an object associated with that
    program which drgn will unwind.

    :param up: a user program, like the one returned by ``get_user_prog()``
    :param data: the bytes of a ``struct pt_regs``, like that returned by
      ``get_pt_regs()``
    """
    # Luckily, all drgn cares about for x86_64 pt_regs is that it is a structure
    # with the right size. Rather than creating a matching struct pt_regs
    # definition, we can just create a dummy one of the correct size:
    #     struct pt_regs {};
    # Drgn will happily use that (not questioning why an empty struct has that
    # size), and we can save ourselves the trouble of creating a convincing
    # replica of the real struct.
    fake_pt_regs_type = up.struct_type(
        tag="pt_regs", size=len(data), members=[]
    )
    return Object.from_bytes_(up, fake_pt_regs_type, data)


def get_tasks(prog: Program, args: argparse.Namespace) -> List[Object]:
    """
    Return the list of tasks according to the dump arguments

    Only task group leaders (i.e. task structs associated with "processes") are
    returned, though this does include kthreads. No task will appear twice in
    the list.
    """
    tasks = []
    task_struct_set = set()

    def add(task: Object) -> None:
        leader = task.group_leader
        if leader.value_() not in task_struct_set:
            task_struct_set.add(leader.value_())
            tasks.append(leader)

    # Only iterate over every task on the system if we have to.
    if args.comm or args.state or args.all:
        comms = [c.encode("utf-8") for c in args.comm]
        for task in for_each_task(prog):
            if task.tgid != task.pid:
                continue  # only handle group leaders

            if args.all:
                add(task)
            elif args.state and task_state_to_char(task) in args.state:
                add(task)
            else:
                comm = task.comm.string_()
                if any(fnmatch.fnmatch(comm, c) for c in comms):
                    add(task)
    # For on-CPU processes, we can efficiently look these up by CPU
    if args.online:
        for cpu in for_each_online_cpu(prog):
            add(cpu_curr(prog, cpu))
    # For PIDs, we can efficiently look these up
    for pid in args.pid:
        add(find_task(prog, pid))
    return tasks


def add_task_args(group: argparse.ArgumentParser) -> None:
    group.add_argument(
        "--online",
        "-o",
        action="store_true",
        help="dump stacks for all on-cpu tasks (not supported for live kernels)",
    )
    group.add_argument(
        "--all",
        "-a",
        action="store_true",
        help="dump stacks for all PIDs (not recommended)",
    )
    group.add_argument(
        "--state",
        "-s",
        action=CommaList,
        default=[],
        help="dump stacks for all tasks in given state (ps(1) 1-letter code)",
    )
    group.add_argument(
        "--comm",
        "-c",
        action=CommaList,
        default=[],
        help="dump stacks for all tasks whose command matches this pattern (glob)",
    )
    group.add_argument(
        "--pid",
        "-p",
        action=CommaList,
        default=[],
        element_type=int,
        help="dump stack for specific PIDs (may be specified multiple times)",
    )


def task_dsos(mm: Object) -> List[Tuple[str, int, int, int]]:
    """
    Return the mapped DSOs for a task's ``mm_struct``. The return value is a
    tuple: (path, start, end, ino)
    """
    # For the first pass, find any mapping which starts at offset 0 within the
    # file, and record the full range of the file (even if it's not actually
    # mapped for its full size). Also, record which file mappings have an
    # executable VMA, since we'll only care about those.
    file_range: Dict[str, Tuple[int, int, int]] = {}
    file_exec = set()
    VM_EXEC = 0x4
    for vma in for_each_vma(mm):
        if not vma.vm_file:
            continue
        path = os.fsdecode(d_path(vma.vm_file.f_path))
        if vma.vm_pgoff == 0:
            file_range[path] = (
                vma.vm_start.value_(),
                (vma.vm_start + vma.vm_file.f_inode.i_size).value_(),
                vma.vm_file.f_inode.i_ino.value_(),
            )
        if vma.vm_flags & VM_EXEC:
            file_exec.add(path)

    # For the second pass, ensure there is no overlap between the ranges we
    # created previously. Since the whole file is not necessarily mapped, we
    # want to avoid creating overlapping ranges that drgn is not expecting (see
    # https://github.com/osandov/drgn/issues/574). Once we've ensured there are
    # no overlaps, and the file had an executable mapping, we can emit it.
    ranges = sorted(file_range.items(), key=lambda t: t[1][0])
    result = []
    for i, (path, (start, end, ino)) in enumerate(ranges):
        if i + 1 < len(ranges):
            end = min(end, ranges[i + 1][1][1])
        if path in file_exec:
            result.append((path, start, end, ino))
    return result


def task_metadata(prog: Program, task: Object) -> Dict[str, Any]:
    """Return JSON metadata about a task, for later use."""
    load_addrs = {}
    if task.mm:
        for path, start, end, inode in task_dsos(task.mm):
            load_addrs[path] = [start, end, inode]
    return {
        "pid": task.pid.value_(),
        "comm": task.comm.string_().decode("utf-8", errors="replace"),
        "mm": load_addrs,
        "kernel": not bool(task.mm),
        "threads": [],
    }


def dump_task(
    task: Object, outfile: gzip.GzipFile, max_stack_bytes: int = 1048576
) -> Tuple[int, int, int]:
    prog = task.prog_
    metadata = task_metadata(prog, task)
    page_mask = ~(prog["PAGE_SIZE"] - 1)
    page_ranges = []
    threads = 0
    for thread in for_each_task_in_group(task, include_self=True):
        tid = thread.pid.value_()
        tcomm = thread.comm.string_().decode("utf-8", errors="replace")
        try:
            kstack = prog.stack_trace(thread)
        except ValueError:
            log.warning("skipped running TID %d ('%s')", tid, tcomm)
            continue
        kstack_str = str(kstack)
        cpu = task_cpu(thread)
        thread_meta = {
            "tid": tid,
            "comm": tcomm,
            "kstack": kstack_str,
            "cpu": cpu,
            "on_cpu": cpu_curr(prog, cpu) == thread,
            "state": task_state_to_char(thread),
        }

        # Kernel threads can still be included even though the value of this
        # tool is getting userspace stacks. Just skip the user registers and
        # stack.
        if not task.mm:
            metadata["threads"].append(thread_meta)
            continue

        # Add user registers to metadata.
        if len(kstack) > 0 and (kstack[0].pc & (1 << 63)):
            # CPU was running in kernel mode, get the saved registers
            regs = task_saved_pt_regs(thread)
        else:
            # CPU was in user mode. Drgn won't make be able to unwind
            # it, but we can take the top frame and get the original
            # registers for unwinding.
            regs = task_running_pt_regs(kstack)
            kstack_str = "<running in user mode>"
        thread_meta["regs"] = base64.b64encode(regs.to_bytes_()).decode()
        metadata["threads"].append(thread_meta)

        # Now get the stack memory range to dump.  Three big assumptions here:
        # (1) the stack grows down, (2) there is no stack switching going on,
        # and (3) the stack is in its own VMA.  These are usually true, but not
        # always.
        vma = vma_find(task.mm, regs.sp)
        if not vma:
            log.warning(
                "could not find VMA for SP (%x) in TID %d ('%s')",
                regs.sp.value_(),
                tid,
                tcomm,
            )
            continue
        start = (regs.sp & page_mask).value_()
        end = vma.vm_end.value_()
        page_ranges.append((start, end))
        threads += 1

    write_json_object(metadata, outfile)

    page_ranges.sort()
    prev_end = 0
    pgsize = prog["PAGE_SIZE"].value_()
    max_stack_bytes = max_stack_bytes & page_mask
    pages_written = pages_faulted = 0
    for start, end in page_ranges:
        # There's no guarantee that we have a reasonable stack size. Apply a
        # limit here to avoid dumping excess data.
        end = min(end, start + max_stack_bytes)

        # It's possible (though unlikely) that the ranges overlap. Avoid
        # re-writing the same pages multiple times.
        start = max(prev_end, start)
        prev_end = end
        for pgaddr in range(start, end, pgsize):
            try:
                data = access_remote_vm(task.mm, pgaddr, pgsize)
            except FaultError:
                pages_faulted += 1
                continue
            outfile.write(struct.pack("=Q", pgaddr))
            outfile.write(data)
            pages_written += 1
    log.info(
        "PID %d: wrote %d pages (%.1f MiB), faulted on %d pages",
        task.pid.value_(),
        pages_written,
        pages_written * pgsize / (1024 * 1024),
        pages_faulted,
    )

    outfile.write(b"\xff" * 8)
    return threads, pages_written, pages_faulted


def write_json_object(obj: Dict[str, Any], outfile: gzip.GzipFile) -> None:
    encoded_data = json.dumps(obj).encode("utf-8")
    outfile.write(struct.pack("=Q", len(encoded_data)))
    outfile.write(encoded_data)


def dump_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "output",
        help="store stack dumps in the given file",
    )
    parser.add_argument(
        "--max-stack-bytes",
        type=int,
        default=(1024 * 1024),
        help="max stack bytes to dump for any thread (default 1MiB)",
    )
    add_task_args(parser)


def dump(prog: Program, args: argparse.Namespace) -> None:
    """
    Write stack information to a compressed binary file.

    The binary format is based on simple blocks of data which are prefixed by an
    8-byte little-endian field which contains either a length, or an address.
    The following blocks are defined and must occur in order.

      - magic header: "pstack" followed by a NUL byte, and a one byte version
      - global metadata: 8-byte length followed by JSON object. The object
        contains "page_size" field, and maybe others. The next block MUST be a
        task metadata.
      - task metadata: 8-byte length followed by JSON object. The object
        represents all tasks sharing the same mm, or for kthreads, a single task
        struct. The object contains an array of threads. The task metadata is
        immediately followed by zero or more stack memory blocks, and then an
        "end of memory" marker.
      - stack memory block: 8-byte field containing a page-aligned address, and
        then one page of data.
      - end of memory block: an 8-byte field containing the signal value
        0xFFFFFFFFFFFFFFFF. The end-of-memory block may be followed by an
        additional task metadata, or the end of file.
    """
    magic = b"pstack\x00\x01"
    file = Path(args.output)
    tasks = threads_written = 0
    pages_written = pages_faulted = 0
    with gzip.open(file, "wb") as f:
        f.write(magic)

        metadata = {
            "page_size": prog["PAGE_SIZE"].value_(),
        }
        write_json_object(metadata, f)

        for task in get_tasks(prog, args):
            threads, written, faulted = dump_task(
                task, f, max_stack_bytes=args.max_stack_bytes
            )
            threads_written += threads
            pages_written += written
            pages_faulted += faulted
            tasks += 1

    print(
        "Dumped {} tasks ({} threads) with {} pages of stack ({:.1f} MiB) -- that's {:.1f} KiB per thread. Skipped {} faulted pages.".format(
            tasks,
            threads_written,
            pages_written,
            (pages_written * prog["PAGE_SIZE"].value_()) / (1024 * 1024),
            (pages_written * prog["PAGE_SIZE"].value_())
            / (1024)
            / threads_written,
            pages_faulted,
        )
    )


def build_prog_from_dump(
    data: List[Tuple[int, bytes]], metadata: Dict[str, Any], page_size: int
) -> Program:
    prog = Program(drgn.host_platform)
    data.sort()

    def read_fn(_, count, offset, __):
        # This may be a bit overkill given that the average single-threaded task
        # only has a few stack pages to speak of. But I can't justify using a
        # linear search when binary search will do better.
        page = offset & ~(page_size - 1)
        # bisect_left gained a "key" kwarg, but in Python 3.10, oh well...
        index = bisect_left(data, (page, b""))
        output = bytearray()
        while len(output) < count:
            if index >= len(data) or data[index][0] != page:
                raise FaultError("memory not present", page)
            pgoff = (offset + len(output)) & (page_size - 1)
            pgend = min(page_size, pgoff + count - len(output))
            output += data[index][1][pgoff:pgend]

            # Move to the next page, which is hopefully present if we have more
            # to read.
            page += page_size
            index += 1
        return output

    prog.add_memory_segment(0, 0xFFFFFFFFFFFFFFFF, read_fn, False)

    pid = metadata["pid"]
    for name, (start, end, ino) in metadata["mm"].items():
        path = Path(name)
        inode = None
        try:
            if path.exists():
                inode = path.stat().st_ino
            else:
                log.warning("For PID %d, could not find file %s", pid, name)
                continue
        except OSError as e:
            log.warning(
                "For PID %d, could not access file %s (%r)", pid, name, e
            )
            continue
        if inode is not None and inode != ino:
            log.warning(
                "For PID %d, file %s inode does not match, file may be updated",
                pid,
                name,
            )
        mod = prog.extra_module(name=name, create=True)
        mod.address_range = (start, end)
        mod.try_file(name, force=True)
    return prog


@lru_cache(maxsize=1)
def _load_demangler() -> Callable[[str], str]:
    # The GNU C++ standard library contains a function called __cxa_demangle
    # which we can use to translate C++ function names.
    #
    # https://gcc.gnu.org/onlinedocs/libstdc++/manual/ext_demangling.html
    # https://gcc.gnu.org/onlinedocs/libstdc++/latest-doxygen/a00026.html#aaf2180d3f67420d4e937e85b281b94a0
    #
    # Signature:
    # char * __cxxabiv1::__cxa_demangle ( const char *__mangled_name,
    #                                     char *__output_buffer,
    #                                     size_t *__length,
    #                                     int *__status
    # )
    # The output buffer, when not provided, is allocated and must be freed via
    # free().
    stdcxx_name = ctypes.util.find_library("stdc++")
    stdc_name = ctypes.util.find_library("c")
    if not stdcxx_name or not stdc_name:
        warnings.warn("Cannot import C++ demangling")
        return lambda s: s
    stdcxx = ctypes.CDLL(stdcxx_name)
    stdc = ctypes.CDLL(stdc_name)
    demangle_func_p = stdcxx.__cxa_demangle
    demangle_func_p.restype = ctypes.POINTER(ctypes.c_char)

    # The status variable has the following return codes:
    status_codes = {
        0: "The demangling operation succeeded.",
        -1: "A memory allocation failure occurred.",
        -2: "mangled_name is not a valid name under the C++ ABI mangling rules.",
        -3: "One of the arguments is invalid.",
    }

    def demanglefn(mangled: str) -> str:
        in_buffer = ctypes.c_char_p(mangled.encode("utf-8"))
        status = ctypes.c_int()
        result = demangle_func_p(in_buffer, None, None, ctypes.pointer(status))
        if status.value != 0:
            msg = status_codes.get(status.value, "unknown error")
            warnings.warn(f"Unable to demangle '{mangled}': {msg}")
            return mangled
        strval = ctypes.cast(result, ctypes.c_char_p).value.decode("utf-8")  # type: ignore
        stdc.free(result)
        return strval

    return demanglefn


def demangle(mangled: str) -> str:
    if mangled.startswith("_Z"):
        return _load_demangler()(mangled)
    else:
        return mangled


def print_user_stack_trace(regs: Object) -> None:
    """
    Prints the userspace stack trace for regs, with the module name included
    for each frame. Including the module name is pretty important for userspace.
    """
    prog = regs.prog_
    trace = prog.stack_trace(regs)
    print("    ------ userspace ---------")
    for i, frame in enumerate(trace):
        name = demangle(frame.name)

        offset = ""
        try:
            sym = frame.symbol()
            offset = f"+0x{frame.pc - sym.address:x}/0x{sym.size:x}"
        except LookupError:
            pass

        mod_text = ""
        try:
            mod = prog.module(frame.pc)
            off = frame.pc - mod.address_range[0]
            mod_text = f" (from {mod.name} +0x{off:x})"
        except LookupError:
            pass

        source_text = ""
        try:
            source_info = ":".join(map(str, frame.source()))
            source_text = f" ({source_info})"
        except LookupError:
            pass

        print(f"    #{i:<2d} {name}{offset}{source_text}{mod_text}")


def dump_print_process(
    fp: gzip.GzipFile, meta: Dict[str, Any], page_size: int
) -> None:
    """Print traces for a dumped process"""
    pid = meta["pid"]
    data = []
    end = b"\xff" * 8
    while True:
        header = fp.read(8)
        if header == end:
            break
        addr = struct.unpack("=Q", header)[0]
        data.append((addr, fp.read(page_size)))

    comm = meta["comm"]
    print(f"[PID: {pid} COMM: {comm}]")

    # Special-case for kernel threads, so we don't build a fake program
    if meta["kernel"]:
        for thread in meta["threads"]:
            print("  " + thread["kstack"].replace("\n", "\n  "))
        return
    prog = build_prog_from_dump(data, meta, page_size)
    for i, t in enumerate(meta["threads"]):
        tid = t["tid"]
        tcomm = t["comm"]
        st = t["state"]
        cpu = t["cpu"]
        cpunote = "RUNNING ON " if t["on_cpu"] else ""
        print(f"  Thread {i} TID={tid} [{st}] {cpunote}CPU={cpu} ('{tcomm}')")
        print("    " + t["kstack"].replace("\n", "\n    "))
        regs = make_fake_pt_regs(prog, base64.b64decode(t["regs"]))
        print_user_stack_trace(regs)


def read_json_object(fp: gzip.GzipFile) -> Dict[str, Any]:
    header = fp.read(8)
    if len(header) != 8:
        raise EOFError("end of file")
    length = struct.unpack("=Q", header)[0]
    return json.loads(fp.read(length))


def dump_print(d: str):
    with gzip.open(d, "rb") as f:
        try:
            magic = f.read(8)
        except OSError:
            sys.exit(f"error: {args.file} doesn't exist or is not a gzip file")
        if magic != b"pstack\x00\x01":
            sys.exit("error: unrecognized file format")

        metadata = read_json_object(f)
        pgsize = metadata["page_size"]

        while True:
            try:
                task_meta = read_json_object(f)
            except EOFError:
                break
            dump_print_process(f, task_meta, pgsize)
            print()


def build_prog_from_mm(mm: Object) -> Program:
    """
    Create a Program representing a userspace task in the kernel Program

    :param mm: the ``struct mm_struct`` for the process
    :returns: a Program which can be debugged like a userspace process
    """
    prog = mm.prog_
    up = Program(prog.platform)

    def read_fn(_, count, offset, __):
        return access_remote_vm(mm, offset, count)

    up.add_memory_segment(0, 0xFFFFFFFFFFFFFFFF, read_fn, False)

    for path, start, end, ino in task_dsos(mm):
        try:
            statbuf = os.stat(path)
            if statbuf.st_ino != ino:
                log.warning(
                    "file %s doesn't match the inode on-disk, it may"
                    " have been updated",
                    path,
                )
        except OSError:
            pass
        mod = up.extra_module(path, create=True)
        mod.address_range = (start, end)
        mod.try_file(path)

    return up


def pstack_print_process(task: Object) -> None:
    comm = task.comm.string_().decode("utf-8", errors="replace")
    print(f"[PID: {task.pid.value_()} COMM: {comm}]")
    prog = task.prog_
    if not task.mm:
        print("  " + str(prog.stack_trace(task)).replace("\n", "\n  "))
        return

    user_prog = build_prog_from_mm(task.mm)
    for i, thread in enumerate(
        for_each_task_in_group(task, include_self=True)
    ):
        tid = thread.pid.value_()
        tcomm = thread.comm.string_().decode("utf-8", errors="replace")
        st = task_state_to_char(thread)
        cpu = task_cpu(thread)
        on_cpu = task_on_cpu(thread)
        cpunote = "RUNNING ON " if on_cpu else ""
        print(f"  Thread {i} TID={tid} [{st}] {cpunote}CPU={cpu} ('{tcomm}')")
        try:
            kstack = prog.stack_trace(thread)
        except ValueError as e:
            if "cannot unwind stack of running task" in str(e):
                print(f"    {str(e)}")
                continue
            else:
                raise
        if len(kstack) > 0 and (kstack[0].pc & (1 << 63)):
            # Kernel stack is indeed a kernel stack, print it
            print(
                "    " + str(prog.stack_trace(thread)).replace("\n", "\n    ")
            )
            regs = task_saved_pt_regs(task)
        else:
            # CPU was in user-mode, print that instead:
            print("    <running in user mode>")
            regs = task_running_pt_regs(kstack)
        fake_regs = make_fake_pt_regs(user_prog, regs.to_bytes_())
        print_user_stack_trace(fake_regs)


def pstack(prog: Program) -> None:
    parser = argparse.ArgumentParser(description="print stack traces")
    add_task_args(parser)
    args = parser.parse_args(sys.argv[2:])
    if args.online and prog.flags & ProgramFlags.IS_LIVE:
        sys.exit("error: --online: cannot unwind running tasks on live system")
    for task in get_tasks(prog, args):
        pstack_print_process(task)
        print()


if __name__ == "__main__":
    logging.basicConfig()
    prog: Program
    if len(sys.argv) <= 1 or sys.argv[1] not in ("dump", "print", "pstack"):
        sys.exit(
            f"usage: drgn [args...] {sys.argv} [dump | print | pstack] ..."
        )
    elif sys.argv[1] == "dump":
        parser = argparse.ArgumentParser(description="dump stacks")
        dump_args(parser)
        dump(prog, parser.parse_args(sys.argv[2:]))  # noqa
    elif sys.argv[1] == "print":
        parser = argparse.ArgumentParser(
            description="print traces for dumped stacks"
        )
        parser.add_argument("file", type=Path, help="compressed")
        args = parser.parse_args(sys.argv[2:])
        dump_print(args.file)
    elif sys.argv[1] == "pstack":
        pstack(prog)  # noqa



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Priority-Sorted Lists
---------------------

The ``drgn.helpers.linux.plist`` module provides helpers for working with
descending-priority-sorted doubly-linked lists (``struct plist_head`` and
``struct plist_node``) from :linux:`include/linux/plist.h`.
"""

from typing import Iterator, Union

from drgn import Object, Type, container_of
from drgn.helpers.linux.list import list_empty, list_for_each_entry

__all__ = (
    "plist_first_entry",
    "plist_for_each",
    "plist_for_each_entry",
    "plist_head_empty",
    "plist_last_entry",
    "plist_node_empty",
)


def plist_head_empty(head: Object) -> bool:
    """
    Return whether a plist is empty.

    :param head: ``struct plist_head *``
    """
    return list_empty(head.node_list.address_of_())


def plist_node_empty(node: Object) -> bool:
    """
    Return whether a plist node is empty (i.e., not on a list).

    :param node: ``struct plist_node *``
    """
    return list_empty(node.node_list.address_of_())


def plist_first_entry(head: Object, type: Union[str, Type], member: str) -> Object:
    """
    Return the first (highest priority) entry in a plist.

    The list is assumed to be non-empty.

    :param head: ``struct plist_head *``
    :param type: Entry type.
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    return container_of(head.node_list.next, type, member + ".node_list")


def plist_last_entry(head: Object, type: Union[str, Type], member: str) -> Object:
    """
    Return the last (lowest priority) entry in a plist.

    The list is assumed to be non-empty.

    :param head: ``struct plist_head *``
    :param type: Entry type.
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    return container_of(head.node_list.prev, type, member + ".node_list")


def plist_for_each(head: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in a plist.

    :param head: ``struct plist_head *``
    :return: Iterator of ``struct plist_node *`` objects.
    """
    return list_for_each_entry(
        "struct plist_node", head.node_list.address_of_(), "node_list"
    )


def plist_for_each_entry(
    type: Union[str, Type], head: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in a plist.

    :param type: Entry type.
    :param head: ``struct plist_head *``
    :param member: Name of plist node member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    return list_for_each_entry(
        type, head.node_list.address_of_(), member + ".node_list"
    )



#!/usr/bin/env drgn
# Copyright (c) Canonical Ltd.
# SPDX-License-Identifier: LGPL-2.1-or-later

""" Script to dump vmallocinfo status using drgn"""

import re
from typing import Optional

from drgn import IntegerLike, Program
from drgn.helpers.linux.mm import for_each_vmap_area

VMAP_RAM = 0x1  # indicates vm_map_ram area

VM_IOREMAP = 0x00000001  # ioremap() and friends
VM_ALLOC = 0x00000002  # vmalloc()
VM_MAP = 0x00000004  # vmap()ed pages
VM_USERMAP = 0x00000008  # suitable for remap_vmalloc_range
VM_DMA_COHERENT = 0x00000010  # dma_alloc_coherent
VM_SPARSE = 0x00001000  # sparse vm_area. not all pages are present.


def is_vmalloc_addr(prog: Program, addr: IntegerLike) -> Optional[bool]:
    vmcoreinfo = prog["VMCOREINFO"].string_()
    match = re.search(
        rb"^NUMBER\(VMALLOC_START\)=(0x[0-9a-f]+)$", vmcoreinfo, flags=re.M
    )
    if match:
        VMALLOC_START = int(match.group(1), 16)

    match = re.search(rb"^NUMBER\(VMALLOC_END\)=(0x[0-9a-f]+)$", vmcoreinfo, flags=re.M)
    if match:
        VMALLOC_END = int(match.group(1), 16)

    try:
        return True if addr >= VMALLOC_START and addr < VMALLOC_END else False
    except:
        return None


for vmap_area in for_each_vmap_area(prog):
    if not vmap_area.vm:
        if vmap_area.flags & VMAP_RAM:
            print(
                f"0x{vmap_area.va_start:x}-0x{vmap_area.va_end:x} {vmap_area.va_end - vmap_area.va_start:10d} vm_map_ram"
            )
            continue
    v = vmap_area.vm
    print(
        f"0x{v.addr.value_():x}-0x{(v.addr+v.size).value_():x} {v.size.value_():10d}",
        end="",
    )
    if v.caller:
        try:
            print(f" {prog.symbol(v.caller.value_()).name}", end="")
        except LookupError:
            print(f" 0x{v.caller.value_():x}", end="")
    if v.nr_pages:
        print(f" pages={v.nr_pages.value_():d}", end="")
    if v.phys_addr:
        print(f" phys=0x{v.phys_addr.value_():x}", end="")
    if v.flags & VM_IOREMAP:
        print(" ioremap", end="")
    if v.flags & VM_SPARSE:
        print(" sparse", end="")
    if v.flags & VM_ALLOC:
        print(" vmalloc", end="")
    if v.flags & VM_MAP:
        print(" vmap", end="")
    if v.flags & VM_USERMAP:
        print(" user", end="")
    if v.flags & VM_DMA_COHERENT:
        print(" dma-coherent", end="")
    if is_vmalloc_addr(prog, v.pages):
        print(" vpages", end="")
    print("")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Locking
-------

The ``drgn.helpers.linux.locking`` module provides helpers for inspecting
locks, including mutexes and read-write semaphores.
"""


import enum

from drgn import Object, Program, TypeKind, cast

__all__ = (
    "mutex_owner",
    "rwsem_locked",
    "rwsem_owner",
)


# Before Linux kernel commit e274795ea7b7 ("locking/mutex: Fix mutex handoff")
# (in v4.11), this was 0x3, but task_struct was sufficiently aligned that it
# doesn't matter.
_MUTEX_FLAGS = 0x7


def mutex_owner(lock: Object) -> Object:
    """
    Get the task that currently owns a mutex.

    Before `Linux 4.10
    <https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=3ca0ff571b092ee4d807f1168caa428d95b0173b>`_,
    this is only supported when ``CONFIG_DEBUG_MUTEXES`` or
    ``CONFIG_MUTEX_SPIN_ON_OWNER`` are enabled.

    :param lock: ``struct mutex *``
    :return: ``struct task_struct *``
    """
    owner = lock.owner
    try:
        # Since Linux kernel commit 3ca0ff571b09 ("locking/mutex: Rework
        # mutex::owner") (in v4.10), struct mutex::owner is a tagged pointer in
        # an atomic_long_t.
        return cast("struct task_struct *", owner.counter & ~_MUTEX_FLAGS)
    except AttributeError:
        # Before that, it was a direct struct task_struct * pointer.
        return owner.read_()


# Linux kernel commit 64489e78004c ("locking/rwsem: Implement a new locking
# scheme") (in v5.3) changed how ownership is tracked. We can't easily detect
# that commit directly, but we can detect a later commit in the same series,
# 94a9717b3c40 ("locking/rwsem: Make rwsem->owner an atomic_long_t").
def _rwsem_new_scheme(prog: Program) -> bool:
    try:
        return prog.cache["rwsem_new_scheme"]
    except KeyError:
        pass
    ret = prog.type("struct rw_semaphore").member("owner").type.kind != TypeKind.POINTER
    prog.cache["rwsem_new_scheme"] = ret
    return ret


def rwsem_locked(sem: Object) -> "RwsemLocked":
    """
    Return whether a read-write semaphore is unlocked, read-locked, or
    write-locked.

    Before `Linux 5.3
    <https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=c71fd893f614f205dbc050d60299cc5496491c19>`_,
    this is only supported when ``CONFIG_RWSEM_SPIN_ON_OWNER`` is enabled.

    :param sem: ``struct rw_semaphore *``
    """
    count = sem.count.counter.value_()
    if not count:
        return RwsemLocked.UNLOCKED
    # Since Linux kernel commit 64489e78004c ("locking/rwsem: Implement a new
    # locking scheme") (in v5.3), we can use the RWSEM_WRITER_LOCKED flag (1)
    # to determine whether a semaphore is write-locked.
    #
    # Before that, count values are ambiguous, so we have to use the owner to
    # disambiguate. Between that and Linux kernel commit 925b9cd1b89a
    # ("locking/rwsem: Make owner store task pointer of last owning reader"),
    # we have to check whether the RWSEM_READER_OWNED flag (1) is set on the
    # owner. Before that, we have to check whether the owner is exactly
    # RWSEM_READER_OWNED (1). Annoyingly, the only way to detect this commit is
    # by the presence of a function.
    if _rwsem_new_scheme(sem.prog_):
        return RwsemLocked.WRITE_LOCKED if (count & 1) else RwsemLocked.READ_LOCKED
    elif "__rwsem_set_reader_owned" in sem.prog_:
        return (
            RwsemLocked.READ_LOCKED
            if sem.owner.value_() & 1
            else RwsemLocked.WRITE_LOCKED
        )
    else:
        return (
            RwsemLocked.READ_LOCKED
            if sem.owner.value_() == 1
            else RwsemLocked.WRITE_LOCKED
        )


class RwsemLocked(enum.Enum):
    """Locked status of a read-write semaphore."""

    UNLOCKED = enum.auto()
    ""
    READ_LOCKED = enum.auto()
    ""
    WRITE_LOCKED = enum.auto()
    ""

    # Make `if rwsem_locked(sem)` work as expected.
    def __bool__(self) -> bool:
        return self != RwsemLocked.UNLOCKED


# Since Linux kernel commit 617f3ef95177 ("locking/rwsem: Remove reader
# optimistic spinning") (in v5.11), this is actually 0x3. Between that and
# 7d43f1ce9dd0 ("locking/rwsem: Enable time-based spinning on reader-owned
# rwsem") (in v5.3), it was 0x7. Between that and 02f1082b003a ("locking/rwsem:
# Clarify usage of owner's nonspinaable bit") (in v5.3), it was also 0x3.
# Between that and 925b9cd1b89a ("locking/rwsem: Make owner store task pointer
# of last owning reader") (in v4.20), the macro name didn't exist, but the mask
# was (RWSEM_READER_OWNED|RWSEM_ANONYMOUSLY_OWNED), which was also 0x3. Before
# that, there was only RWSEM_ANONYMOUSLY_OWNED (0x1). task_struct is
# sufficiently aligned that none of this matters and we can just use 0x7.
_RWSEM_OWNER_FLAGS_MASK = 0x7


def rwsem_owner(sem: Object) -> Object:
    """
    Get the task that currently owns a read-write semaphore.

    .. warning::

        Due to the kernel implementation, unless the semaphore is currently
        write-locked, then this is not totally reliable.

        It may return ``NULL`` when the semaphore is read-locked (specifically,
        when the last task to read-lock the semaphore unlocks it and
        ``CONFIG_DEBUG_RWSEMS`` or ``CONFIG_DETECT_HUNG_TASK_BLOCKER`` are
        enabled, or before `Linux 4.20
        <https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=925b9cd1b89a94b7124d128c80dfc48f78a63098>`_).

        It may also return a previous reader that no longer owns the semaphore
        when it is read-locked or unlocked (although an effort is made to avoid
        this when ``CONFIG_DEBUG_RWSEMS`` or
        ``CONFIG_DETECT_HUNG_TASK_BLOCKER`` are enabled unless
        ``up_read_non_owner()`` is used).

    Before `Linux 5.3
    <https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=c71fd893f614f205dbc050d60299cc5496491c19>`_,
    this is only supported when ``CONFIG_RWSEM_SPIN_ON_OWNER`` is enabled.

    :param sem: ``struct rw_semaphore *``
    :return: ``struct task_struct *``
    """
    # Since Linux kernel commit 94a9717b3c40 ("locking/rwsem: Make rwsem->owner
    # an atomic_long_t") (in v5.3), struct rw_semaphore::owner is an
    # atomic_long_t. Before that, it was a struct task_struct *. On all
    # kernels, it's actually a tagged pointer.
    if _rwsem_new_scheme(sem.prog_):
        counter = sem.owner.counter.value_()
    else:
        counter = sem.owner.value_()
    return Object(sem.prog_, "struct task_struct *", counter & ~_RWSEM_OWNER_FLAGS_MASK)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import platform
import re
import sys
import types

NORMALIZED_MACHINE_NAME = platform.machine()
if NORMALIZED_MACHINE_NAME.startswith("aarch64") or NORMALIZED_MACHINE_NAME == "arm64":
    NORMALIZED_MACHINE_NAME = "aarch64"
elif NORMALIZED_MACHINE_NAME.startswith("arm") or NORMALIZED_MACHINE_NAME == "sa110":
    NORMALIZED_MACHINE_NAME = "arm"
elif re.fullmatch(r"i.86", NORMALIZED_MACHINE_NAME):
    NORMALIZED_MACHINE_NAME = "i386"
elif NORMALIZED_MACHINE_NAME.startswith("ppc64"):
    NORMALIZED_MACHINE_NAME = "ppc64"
elif NORMALIZED_MACHINE_NAME.startswith("ppc"):
    NORMALIZED_MACHINE_NAME = "ppc"
elif NORMALIZED_MACHINE_NAME == "riscv":
    NORMALIZED_MACHINE_NAME = "riscv32"
elif re.match(r"sh[0-9]", NORMALIZED_MACHINE_NAME):
    NORMALIZED_MACHINE_NAME = "sh"
elif NORMALIZED_MACHINE_NAME == "sun4u":
    NORMALIZED_MACHINE_NAME = "sparc64"


if NORMALIZED_MACHINE_NAME == "x86_64":
    if sys.maxsize > 2**32:
        SYS = {
            "bpf": 321,
            "finit_module": 313,
            "io_destroy": 207,
            "io_setup": 206,
            "io_submit": 209,
            "kexec_file_load": 320,
            "memfd_create": 319,
            "perf_event_open": 298,
        }
    else:  # x32
        SYS = {
            "bpf": 321,
            "finit_module": 313,
            "io_destroy": 207,
            "io_setup": 543,
            "io_submit": 544,
            "kexec_file_load": 320,
            "memfd_create": 319,
            "perf_event_open": 298,
        }
else:
    SYS = {
        "aarch64": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "alpha": {
            "bpf": 515,
            "finit_module": 507,
            "io_destroy": 399,
            "io_setup": 398,
            "io_submit": 401,
            "memfd_create": 512,
            "perf_event_open": 493,
        },
        "arc": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "arm": {
            "bpf": 386,
            "finit_module": 379,
            "io_destroy": 244,
            "io_setup": 243,
            "io_submit": 246,
            "kexec_file_load": 401,
            "memfd_create": 385,
            "perf_event_open": 364,
        },
        "csky": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "hexagon": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "i386": {
            "bpf": 357,
            "finit_module": 350,
            "io_destroy": 246,
            "io_setup": 245,
            "io_submit": 248,
            "memfd_create": 356,
            "perf_event_open": 336,
        },
        "loongarch": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "loongarch64": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "m68k": {
            "bpf": 354,
            "finit_module": 348,
            "io_destroy": 242,
            "io_setup": 241,
            "io_submit": 244,
            "memfd_create": 353,
            "perf_event_open": 332,
        },
        "microblaze": {
            "bpf": 387,
            "finit_module": 380,
            "io_destroy": 246,
            "io_setup": 245,
            "io_submit": 248,
            "memfd_create": 386,
            "perf_event_open": 366,
        },
        # TODO: mips is missing here because I don't know how to distinguish
        # between the o32 and n32 ABIs.
        "mips64": {
            "bpf": 315,
            "finit_module": 307,
            "io_destroy": 201,
            "io_setup": 200,
            "io_submit": 203,
            "memfd_create": 314,
            "perf_event_open": 292,
        },
        "nios2": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "openrisc": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "parisc": {
            "bpf": 341,
            "finit_module": 333,
            "io_destroy": 216,
            "io_setup": 215,
            "io_submit": 218,
            "kexec_file_load": 355,
            "memfd_create": 340,
            "perf_event_open": 318,
        },
        "parisc64": {
            "bpf": 341,
            "finit_module": 333,
            "io_destroy": 216,
            "io_setup": 215,
            "io_submit": 218,
            "kexec_file_load": 355,
            "memfd_create": 340,
            "perf_event_open": 318,
        },
        "ppc": {
            "bpf": 361,
            "finit_module": 353,
            "io_destroy": 228,
            "io_setup": 227,
            "io_submit": 230,
            "memfd_create": 360,
            "perf_event_open": 319,
        },
        "ppc64": {
            "bpf": 361,
            "finit_module": 353,
            "io_destroy": 228,
            "io_setup": 227,
            "io_submit": 230,
            "memfd_create": 360,
            "perf_event_open": 319,
        },
        "riscv32": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "riscv64": {
            "bpf": 280,
            "finit_module": 273,
            "io_destroy": 1,
            "io_setup": 0,
            "io_submit": 2,
            "kexec_file_load": 294,
            "memfd_create": 279,
            "perf_event_open": 241,
        },
        "s390": {
            "bpf": 351,
            "finit_module": 344,
            "io_destroy": 244,
            "io_setup": 243,
            "io_submit": 246,
            "kexec_file_load": 381,
            "memfd_create": 350,
            "perf_event_open": 331,
        },
        "s390x": {
            "bpf": 351,
            "finit_module": 344,
            "io_destroy": 244,
            "io_setup": 243,
            "io_submit": 246,
            "kexec_file_load": 381,
            "memfd_create": 350,
            "perf_event_open": 331,
        },
        "sh": {
            "bpf": 375,
            "finit_module": 368,
            "io_destroy": 246,
            "io_setup": 245,
            "io_submit": 248,
            "memfd_create": 374,
            "perf_event_open": 336,
        },
        "sparc": {
            "bpf": 349,
            "finit_module": 342,
            "io_destroy": 269,
            "io_setup": 268,
            "io_submit": 270,
            "memfd_create": 348,
            "perf_event_open": 327,
        },
        "sparc64": {
            "bpf": 349,
            "finit_module": 342,
            "io_destroy": 269,
            "io_setup": 268,
            "io_submit": 270,
            "memfd_create": 348,
            "perf_event_open": 327,
        },
        "xtensa": {
            "bpf": 340,
            "finit_module": 332,
            "io_destroy": 240,
            "io_setup": 239,
            "io_submit": 241,
            "memfd_create": 339,
            "perf_event_open": 327,
        },
    }.get(NORMALIZED_MACHINE_NAME, {})

_IOC_NONE = {
    "alpha": 1,
    "mips": 1,
    "mips64": 1,
    "ppc": 1,
    "ppc64": 1,
    "sparc": 1,
    "sparc64": 1,
}.get(NORMALIZED_MACHINE_NAME, 0)

_IOC_READ = {
    "parisc": 1,
    "parisc64": 1,
}.get(NORMALIZED_MACHINE_NAME, 2)

_IOC_WRITE = {
    "alpha": 4,
    "mips": 4,
    "mips64": 4,
    "parisc": 2,
    "parisc64": 2,
    "ppc": 4,
    "ppc64": 4,
    "sparc": 4,
    "sparc64": 4,
}.get(NORMALIZED_MACHINE_NAME, 1)

_IOC_NRBITS = 8

_IOC_TYPEBITS = 8

_IOC_SIZEBITS = {
    "alpha": 13,
    "mips": 13,
    "mips64": 13,
    "ppc": 13,
    "ppc64": 13,
    "sparc": 13,
    "sparc64": 13,
}.get(NORMALIZED_MACHINE_NAME, 14)

_IOC_DIRBITS = {
    "alpha": 3,
    "mips": 3,
    "mips64": 3,
    "ppc": 3,
    "ppc64": 3,
    "sparc": 3,
    "sparc64": 3,
}.get(NORMALIZED_MACHINE_NAME, 2)

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS


def _IOC(dir: int, type: int, nr: int, size: int) -> int:
    return (
        (dir << _IOC_DIRSHIFT)
        | (type << _IOC_TYPESHIFT)
        | (nr << _IOC_NRSHIFT)
        | (size << _IOC_SIZESHIFT)
    )


def _IO(type: int, nr: int) -> int:
    return _IOC(_IOC_NONE, type, nr, 0)


def _IOR(type: int, nr: int, size: int) -> int:
    return _IOC(_IOC_READ, type, nr, size)


def _IOW(type: int, nr: int, size: int) -> int:
    return _IOC(_IOC_WRITE, type, nr, size)


def _IOWR(type: int, nr: int, size: int) -> int:
    return _IOC(_IOC_READ | _IOC_WRITE, type, nr, size)


_SIGNALS_BY_MACHINE_NAME_TMP = {}
for _name in (
    "aarch64",
    "arc",
    "csky",
    "hexagon",
    "i386",
    "loongarch",
    "loongarch64",
    "m68k",
    "microblaze",
    "nios2",
    "openrisc",
    "ppc",
    "ppc64",
    "riscv32",
    "riscv64",
    "s390",
    "s390x",
    "sh",
    "x86_64",
):
    _SIGNALS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SIGHUP": 1,
            "SIGINT": 2,
            "SIGQUIT": 3,
            "SIGILL": 4,
            "SIGTRAP": 5,
            "SIGABRT": 6,
            "SIGIOT": 6,
            "SIGBUS": 7,
            "SIGFPE": 8,
            "SIGKILL": 9,
            "SIGUSR1": 10,
            "SIGSEGV": 11,
            "SIGUSR2": 12,
            "SIGPIPE": 13,
            "SIGALRM": 14,
            "SIGTERM": 15,
            "SIGSTKFLT": 16,
            "SIGCHLD": 17,
            "SIGCLD": 17,
            "SIGCONT": 18,
            "SIGSTOP": 19,
            "SIGTSTP": 20,
            "SIGTTIN": 21,
            "SIGTTOU": 22,
            "SIGURG": 23,
            "SIGXCPU": 24,
            "SIGXFSZ": 25,
            "SIGVTALRM": 26,
            "SIGPROF": 27,
            "SIGWINCH": 28,
            "SIGIO": 29,
            "SIGPOLL": 29,
            "SIGPWR": 30,
            "SIGSYS": 31,
            "SIGUNUSED": 31,
            "SIGRTMIN": 32,
            "SIGRTMAX": 64,
        }
    )
_SIGNALS_BY_MACHINE_NAME_TMP["alpha"] = types.MappingProxyType(
    {
        "SIGHUP": 1,
        "SIGINT": 2,
        "SIGQUIT": 3,
        "SIGILL": 4,
        "SIGTRAP": 5,
        "SIGABRT": 6,
        "SIGIOT": 6,
        "SIGEMT": 7,
        "SIGFPE": 8,
        "SIGKILL": 9,
        "SIGBUS": 10,
        "SIGSEGV": 11,
        "SIGSYS": 12,
        "SIGUNUSED": 12,
        "SIGPIPE": 13,
        "SIGALRM": 14,
        "SIGTERM": 15,
        "SIGURG": 16,
        "SIGSTOP": 17,
        "SIGTSTP": 18,
        "SIGCONT": 19,
        "SIGCHLD": 20,
        "SIGCLD": 20,
        "SIGTTIN": 21,
        "SIGTTOU": 22,
        "SIGIO": 23,
        "SIGPOLL": 23,
        "SIGXCPU": 24,
        "SIGXFSZ": 25,
        "SIGVTALRM": 26,
        "SIGPROF": 27,
        "SIGWINCH": 28,
        "SIGPWR": 29,
        "SIGINFO": 29,
        "SIGUSR1": 30,
        "SIGUSR2": 31,
        "SIGRTMIN": 32,
        "SIGRTMAX": 64,
    }
)
_SIGNALS_BY_MACHINE_NAME_TMP["arm"] = types.MappingProxyType(
    {
        "SIGHUP": 1,
        "SIGINT": 2,
        "SIGQUIT": 3,
        "SIGILL": 4,
        "SIGTRAP": 5,
        "SIGABRT": 6,
        "SIGIOT": 6,
        "SIGBUS": 7,
        "SIGFPE": 8,
        "SIGKILL": 9,
        "SIGUSR1": 10,
        "SIGSEGV": 11,
        "SIGUSR2": 12,
        "SIGPIPE": 13,
        "SIGALRM": 14,
        "SIGTERM": 15,
        "SIGSTKFLT": 16,
        "SIGCHLD": 17,
        "SIGCLD": 17,
        "SIGCONT": 18,
        "SIGSTOP": 19,
        "SIGTSTP": 20,
        "SIGTTIN": 21,
        "SIGTTOU": 22,
        "SIGURG": 23,
        "SIGXCPU": 24,
        "SIGXFSZ": 25,
        "SIGVTALRM": 26,
        "SIGPROF": 27,
        "SIGWINCH": 28,
        "SIGIO": 29,
        "SIGPOLL": 29,
        "SIGPWR": 30,
        "SIGSYS": 31,
        "SIGUNUSED": 31,
        "SIGRTMIN": 32,
        "SIGSWI": 32,
        "SIGRTMAX": 64,
    }
)
for _name in ("mips", "mips64"):
    _SIGNALS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SIGHUP": 1,
            "SIGINT": 2,
            "SIGQUIT": 3,
            "SIGILL": 4,
            "SIGTRAP": 5,
            "SIGABRT": 6,
            "SIGIOT": 6,
            "SIGEMT": 7,
            "SIGFPE": 8,
            "SIGKILL": 9,
            "SIGBUS": 10,
            "SIGSEGV": 11,
            "SIGSYS": 12,
            "SIGUNUSED": 12,
            "SIGPIPE": 13,
            "SIGALRM": 14,
            "SIGTERM": 15,
            "SIGUSR1": 16,
            "SIGUSR2": 17,
            "SIGCHLD": 18,
            "SIGCLD": 18,
            "SIGPWR": 19,
            "SIGWINCH": 20,
            "SIGURG": 21,
            "SIGIO": 22,
            "SIGPOLL": 22,
            "SIGSTOP": 23,
            "SIGTSTP": 24,
            "SIGCONT": 25,
            "SIGTTIN": 26,
            "SIGTTOU": 27,
            "SIGVTALRM": 28,
            "SIGPROF": 29,
            "SIGXCPU": 30,
            "SIGXFSZ": 31,
            "SIGRTMIN": 32,
            "SIGRTMAX": 128,
        }
    )
for _name in ("parisc", "parisc64"):
    _SIGNALS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SIGHUP": 1,
            "SIGINT": 2,
            "SIGQUIT": 3,
            "SIGILL": 4,
            "SIGTRAP": 5,
            "SIGABRT": 6,
            "SIGIOT": 6,
            "SIGSTKFLT": 7,
            "SIGFPE": 8,
            "SIGKILL": 9,
            "SIGBUS": 10,
            "SIGSEGV": 11,
            "SIGXCPU": 12,
            "SIGPIPE": 13,
            "SIGALRM": 14,
            "SIGTERM": 15,
            "SIGUSR1": 16,
            "SIGUSR2": 17,
            "SIGCHLD": 18,
            "SIGCLD": 18,
            "SIGPWR": 19,
            "SIGVTALRM": 20,
            "SIGPROF": 21,
            "SIGIO": 22,
            "SIGPOLL": 22,
            "SIGWINCH": 23,
            "SIGSTOP": 24,
            "SIGTSTP": 25,
            "SIGCONT": 26,
            "SIGTTIN": 27,
            "SIGTTOU": 28,
            "SIGURG": 29,
            "SIGXFSZ": 30,
            "SIGSYS": 31,
            "SIGUNUSED": 31,
            "SIGRTMIN": 32,
            "SIGRTMAX": 64,
        }
    )
for _name in ("sparc", "sparc64"):
    _SIGNALS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SIGHUP": 1,
            "SIGINT": 2,
            "SIGQUIT": 3,
            "SIGILL": 4,
            "SIGTRAP": 5,
            "SIGABRT": 6,
            "SIGIOT": 6,
            "SIGEMT": 7,
            "SIGFPE": 8,
            "SIGKILL": 9,
            "SIGBUS": 10,
            "SIGSEGV": 11,
            "SIGSYS": 12,
            "SIGUNUSED": 12,
            "SIGPIPE": 13,
            "SIGALRM": 14,
            "SIGTERM": 15,
            "SIGURG": 16,
            "SIGSTOP": 17,
            "SIGTSTP": 18,
            "SIGCONT": 19,
            "SIGCHLD": 20,
            "SIGCLD": 20,
            "SIGTTIN": 21,
            "SIGTTOU": 22,
            "SIGIO": 23,
            "SIGPOLL": 23,
            "SIGXCPU": 24,
            "SIGXFSZ": 25,
            "SIGVTALRM": 26,
            "SIGPROF": 27,
            "SIGWINCH": 28,
            "SIGPWR": 29,
            "SIGLOST": 29,
            "SIGUSR1": 30,
            "SIGUSR2": 31,
            "SIGRTMIN": 32,
            "SIGRTMAX": 64,
        }
    )
_SIGNALS_BY_MACHINE_NAME_TMP["xtensa"] = types.MappingProxyType(
    {
        "SIGHUP": 1,
        "SIGINT": 2,
        "SIGQUIT": 3,
        "SIGILL": 4,
        "SIGTRAP": 5,
        "SIGABRT": 6,
        "SIGIOT": 6,
        "SIGBUS": 7,
        "SIGFPE": 8,
        "SIGKILL": 9,
        "SIGUSR1": 10,
        "SIGSEGV": 11,
        "SIGUSR2": 12,
        "SIGPIPE": 13,
        "SIGALRM": 14,
        "SIGTERM": 15,
        "SIGSTKFLT": 16,
        "SIGCHLD": 17,
        "SIGCLD": 17,
        "SIGCONT": 18,
        "SIGSTOP": 19,
        "SIGTSTP": 20,
        "SIGTTIN": 21,
        "SIGTTOU": 22,
        "SIGURG": 23,
        "SIGXCPU": 24,
        "SIGXFSZ": 25,
        "SIGVTALRM": 26,
        "SIGPROF": 27,
        "SIGWINCH": 28,
        "SIGIO": 29,
        "SIGPOLL": 29,
        "SIGPWR": 30,
        "SIGSYS": 31,
        "SIGUNUSED": 31,
        "SIGRTMIN": 32,
        "SIGRTMAX": 63,
    }
)
SIGNALS_BY_MACHINE_NAME = types.MappingProxyType(_SIGNALS_BY_MACHINE_NAME_TMP)
del _SIGNALS_BY_MACHINE_NAME_TMP

_SIGACTION_FLAGS_BY_MACHINE_NAME_TMP = {}
for _name in (
    "aarch64",
    "arc",
    "nios2",
    "ppc",
    "ppc64",
    "s390",
    "s390x",
    "sh",
    "xtensa",
):
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SA_NOCLDSTOP": 0x1,
            "SA_NOCLDWAIT": 0x2,
            "SA_SIGINFO": 0x4,
            "SA_UNSUPPORTED": 0x400,
            "SA_EXPOSE_TAGBITS": 0x800,
            "SA_IMMUTABLE": 0x800000,
            "SA_RESTORER": 0x4000000,
            "SA_ONSTACK": 0x8000000,
            "SA_STACK": 0x8000000,
            "SA_RESTART": 0x10000000,
            "SA_NODEFER": 0x40000000,
            "SA_NOMASK": 0x40000000,
            "SA_RESETHAND": 0x80000000,
            "SA_ONESHOT": 0x80000000,
        }
    )
_SIGACTION_FLAGS_BY_MACHINE_NAME_TMP["alpha"] = types.MappingProxyType(
    {
        "SA_ONSTACK": 0x1,
        "SA_STACK": 0x1,
        "SA_RESTART": 0x2,
        "SA_NOCLDSTOP": 0x4,
        "SA_NODEFER": 0x8,
        "SA_NOMASK": 0x8,
        "SA_RESETHAND": 0x10,
        "SA_ONESHOT": 0x10,
        "SA_NOCLDWAIT": 0x20,
        "SA_SIGINFO": 0x40,
        "SA_UNSUPPORTED": 0x400,
        "SA_EXPOSE_TAGBITS": 0x800,
        "SA_IMMUTABLE": 0x800000,
    }
)
_SIGACTION_FLAGS_BY_MACHINE_NAME_TMP["arm"] = types.MappingProxyType(
    {
        "SA_NOCLDSTOP": 0x1,
        "SA_NOCLDWAIT": 0x2,
        "SA_SIGINFO": 0x4,
        "SA_UNSUPPORTED": 0x400,
        "SA_EXPOSE_TAGBITS": 0x800,
        "SA_IMMUTABLE": 0x800000,
        "SA_THIRTYTWO": 0x2000000,
        "SA_RESTORER": 0x4000000,
        "SA_ONSTACK": 0x8000000,
        "SA_STACK": 0x8000000,
        "SA_RESTART": 0x10000000,
        "SA_NODEFER": 0x40000000,
        "SA_NOMASK": 0x40000000,
        "SA_RESETHAND": 0x80000000,
        "SA_ONESHOT": 0x80000000,
    }
)
for _name in (
    "csky",
    "hexagon",
    "loongarch",
    "loongarch64",
    "m68k",
    "microblaze",
    "openrisc",
    "riscv32",
    "riscv64",
):
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SA_NOCLDSTOP": 0x1,
            "SA_NOCLDWAIT": 0x2,
            "SA_SIGINFO": 0x4,
            "SA_UNSUPPORTED": 0x400,
            "SA_EXPOSE_TAGBITS": 0x800,
            "SA_IMMUTABLE": 0x800000,
            "SA_ONSTACK": 0x8000000,
            "SA_STACK": 0x8000000,
            "SA_RESTART": 0x10000000,
            "SA_NODEFER": 0x40000000,
            "SA_NOMASK": 0x40000000,
            "SA_RESETHAND": 0x80000000,
            "SA_ONESHOT": 0x80000000,
        }
    )
for _name in ("i386", "x86_64"):
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SA_NOCLDSTOP": 0x1,
            "SA_NOCLDWAIT": 0x2,
            "SA_SIGINFO": 0x4,
            "SA_UNSUPPORTED": 0x400,
            "SA_EXPOSE_TAGBITS": 0x800,
            "SA_IMMUTABLE": 0x800000,
            "SA_X32_ABI": 0x1000000,
            "SA_IA32_ABI": 0x2000000,
            "SA_RESTORER": 0x4000000,
            "SA_ONSTACK": 0x8000000,
            "SA_STACK": 0x8000000,
            "SA_RESTART": 0x10000000,
            "SA_NODEFER": 0x40000000,
            "SA_NOMASK": 0x40000000,
            "SA_RESETHAND": 0x80000000,
            "SA_ONESHOT": 0x80000000,
        }
    )
for _name in ("mips", "mips64"):
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SA_NOCLDSTOP": 0x1,
            "SA_SIGINFO": 0x8,
            "SA_UNSUPPORTED": 0x400,
            "SA_EXPOSE_TAGBITS": 0x800,
            "SA_NOCLDWAIT": 0x10000,
            "SA_IMMUTABLE": 0x800000,
            "SA_ONSTACK": 0x8000000,
            "SA_STACK": 0x8000000,
            "SA_RESTART": 0x10000000,
            "SA_NODEFER": 0x40000000,
            "SA_NOMASK": 0x40000000,
            "SA_RESETHAND": 0x80000000,
            "SA_ONESHOT": 0x80000000,
        }
    )
for _name in ("parisc", "parisc64"):
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SA_ONSTACK": 0x1,
            "SA_STACK": 0x1,
            "SA_RESETHAND": 0x4,
            "SA_ONESHOT": 0x4,
            "SA_NOCLDSTOP": 0x8,
            "SA_SIGINFO": 0x10,
            "SA_NODEFER": 0x20,
            "SA_NOMASK": 0x20,
            "SA_RESTART": 0x40,
            "SA_NOCLDWAIT": 0x80,
            "SA_UNSUPPORTED": 0x400,
            "SA_EXPOSE_TAGBITS": 0x800,
            "SA_IMMUTABLE": 0x800000,
        }
    )
for _name in ("sparc", "sparc64"):
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP[_name] = types.MappingProxyType(
        {
            "SA_ONSTACK": 0x1,
            "SA_STACK": 0x1,
            "SA_RESTART": 0x2,
            "SA_RESETHAND": 0x4,
            "SA_ONESHOT": 0x4,
            "SA_NOCLDSTOP": 0x8,
            "SA_NODEFER": 0x20,
            "SA_NOMASK": 0x20,
            "SA_NOCLDWAIT": 0x100,
            "SA_SIGINFO": 0x200,
            "SA_UNSUPPORTED": 0x400,
            "SA_EXPOSE_TAGBITS": 0x800,
            "SA_IMMUTABLE": 0x800000,
        }
    )
SIGACTION_FLAGS_BY_MACHINE_NAME = types.MappingProxyType(
    _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP
)
del _SIGACTION_FLAGS_BY_MACHINE_NAME_TMP



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, Union, cast

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import Concatenate, ParamSpec
    else:
        from typing import Concatenate, ParamSpec  # novermin

    P = ParamSpec("P")

    class _Method(Protocol[P]):
        def __call__(s, /, self: Any, *args: P.args, **kwargs: P.kwargs) -> Any: ...


A = TypeVar("A")
T = TypeVar("T")
C = TypeVar("C", bound=Callable[..., Any])


# Based on https://github.com/python/cpython/pull/121693.
def copy_func_params(
    source: Callable[P, Any]
) -> Callable[[Callable[..., T]], Callable[P, T]]:
    def decorator(func: Callable[..., T]) -> Callable[P, T]:
        return cast("Callable[P, T]", func)

    return decorator


def copy_func_signature(source: C) -> Callable[[Callable[..., Any]], C]:
    def decorator(func: Callable[..., Any]) -> C:
        return cast(C, func)

    return decorator


def copy_method_params(
    source: Union[
        Callable[Concatenate[Any, P], Any],
        # This covers the case where the method has non-positional-only self
        # and **kwargs. See python/mypy#19051.
        _Method[P],
    ]
) -> Callable[[Callable[Concatenate[A, ...], T]], Callable[Concatenate[A, P], T]]:
    def decorator(
        # For some reason, the same _Method trick we use for source doesn't
        # work here, so if the target takes **kwargs, self must be
        # positional-only.
        func: Callable[Concatenate[Any, ...], T]
    ) -> Callable[Concatenate[A, P], T]:
        return cast("Callable[Concatenate[A, P], T]", func)

    return decorator



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Users
-----

The ``drgn.helpers.linux.user`` module provides helpers for working with users
in the Linux kernel.
"""

import operator
from typing import Iterator, Union

from drgn import NULL, IntegerLike, Object, Program, TypeKind
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.list import hlist_for_each_entry

__all__ = (
    "find_user",
    "for_each_user",
    "kuid_val",
)


def kuid_val(uid: Union[Object, IntegerLike]) -> int:
    """
    Return the UID value of a ``kuid_t``.

    :param uid: ``kuid_t`` object. May also be an integer, in which case its
        value is returned.
    """
    if isinstance(uid, Object) and uid.type_.unaliased_kind() != TypeKind.INT:
        uid = uid.val
    return operator.index(uid)


@takes_program_or_default
def find_user(prog: Program, uid: Union[Object, IntegerLike]) -> Object:
    """
    Return the user structure with the given UID.

    :param uid: ``kuid_t`` object or integer.
    :return: ``struct user_struct *`` (``NULL`` if not found)
    """
    try:
        uidhashentry = prog.cache["uidhashentry"]
    except KeyError:
        uidhash_table = prog["uidhash_table"]
        uidhash_sz = len(uidhash_table)
        uidhash_bits = uidhash_sz.bit_length() - 1
        uidhash_mask = uidhash_sz - 1

        def uidhashentry(uid: int) -> Object:
            hash = ((uid >> uidhash_bits) + uid) & uidhash_mask
            return uidhash_table + hash

        prog.cache["uidhashentry"] = uidhashentry

    uid = kuid_val(uid)
    for user in hlist_for_each_entry(
        "struct user_struct", uidhashentry(uid), "uidhash_node"
    ):
        if user.uid.val == uid:
            return user
    return NULL(prog, "struct user_struct *")


@takes_program_or_default
def for_each_user(prog: Program) -> Iterator[Object]:
    """
    Iterate over all users in the system.

    :return: Iterator of ``struct user_struct *`` objects.
    """
    for hash_entry in prog["uidhash_table"]:
        yield from hlist_for_each_entry(
            "struct user_struct", hash_entry, "uidhash_node"
        )
